class "GtkWidget" style:highest "default"
class "GtkScrollbar" style:highest "ardour_adjusters"
class "GtkLabel" style:highest "default_generic"
class "GtkButton" style:highest "ardour_button"
class "GtkCheckButton" style:highest "ardour_button"
class "GtkArrow" style:highest "tearoff_arrow"
class "GtkProgressBar" style:highest "ardour_progressbars"

widget "gtk-tooltip*" style:highest "tooltip"

widget "*PaddedButton" style:highest "padded_button"
widget "*VerboseCanvasCursor" style:highest "verbose_canvas_cursor"
widget "*MarkerText" style:highest "marker_text"
widget "*TimeAxisViewItemName*" style:highest "time_axis_view_item_name"
widget "*EditModeSelector" style:highest "medium_bold_entry"
widget "*SnapTypeSelector" style:highest "medium_bold_entry"
widget "*SnapModeSelector" style:highest "medium_bold_entry"
widget "*ZoomFocusSelector" style:highest "medium_bold_entry"
widget "*ArdourContextMenu*" style:highest "default_menu"
widget "*EditGroupTitleButton*" style:highest "default_button"
widget "*MixerGroupTitleButton*" style:highest "default_button"
widget "*ErrorLogCloseButton" style:highest "default_button"
widget "*EditorGTKButton*" style:highest "default_button" 
widget "*ToolbarButton" style:highest "default_button"
widget "*ToolbarButton*" style:highest "default_button"
widget "*CrossfadeEditButton" style:highest "default_button"
widget "*CrossfadeEditButton*" style:highest "default_button"
widget "*TrackHistoryButton*" style:highest "default_button"
widget "*TrackSizeButton*" style:highest "default_button"
widget "*TrackPlaylistButton*" style:highest "default_button"
widget "*TrackAutomationButton*" style:highest "default_button"
widget "*TrackGroupButton*" style:highest "default_button"
widget "*TrackMixButton*" style:highest "default_button"
widget "*TrackVisualButton*" style:highest "default_button"
widget "*TrackRemoveButton*" style:highest "default_button"
widget "*BaseButton" style:highest "default_button"
widget "*TakeButtonLabel" style:highest "default_button"

widget "*mixer strip button" style:highest "very_small_button"

widget "*AddRouteDialogSpinner" style:highest "ardour_adjusters"
widget "*AddRouteDialogRadioButton*" style:highest "preferences"
widget "*OptionsNotebook" style:highest "preferences"
widget "*OptionEditorToggleButton*" style:highest "preferences"
widget "*OptionsLabel" style:highest "preferences"
widget "*OptionEditorAuditionerLabel" style:highest "preferences"
widget "*OptionsEntry" style:highest "option_entry"
widget "*InspectorNotebook" style:highest "preferences"
widget "*NewSessionDialog" style:highest "preferences"
widget "*NewSessionDialogButton*" style:highest "preferences"
widget "*MixerSendSwitch*" style:highest "very_small_red_active_and_selected_button"
widget "*OptionEditorToggleButton" style:highest "small_red_active_and_selected_button"
widget "*NewSessionDialogButton" style:highest "small_red_active_and_selected_button"
widget "*RecordEnableButton" style:highest "track_rec_enable_button"
widget "*RecordEnableButton-active" style:highest "track_rec_enable_button_active"
widget "*RecordEnableButton-alternate" style:highest "track_rec_enable_button_alternate"
widget "*RecordEnableButton-alternate2" style:highest "track_rec_enable_button_alternate_two"
widget "*MixerRecordEnableButton" style:highest "mixer_track_rec_enable_button"
widget "*MixerRecordEnableButton-active" style:highest "mixer_track_rec_enable_button_active"
widget "*MixerRecordEnableButton-alternate" style:highest "mixer_track_rec_enable_button_alternate"
widget "*MuteButton" style:highest "mute_button"
widget "*MuteButton-alternate" style:highest "mute_button_fullylit"
widget "*MuteButton-active" style:highest "mute_button_halflit"
widget "*MixerMuteButton" style:highest "mixer_mute_button"
widget "*MixerMuteButton-alternate" style:highest "mixer_mute_button_fullylit"
widget "*MixerMuteButton-active" style:highest "mixer_mute_button_halflit"
widget "*SoloButton" style:highest "solo_button"
widget "*SoloButton-alternate" style:highest "solo_button_alternate"
widget "*SoloButton-alternate2" style:highest "solo_button_alternate2"
widget "*SoloButton-active" style:highest "solo_button_active"
widget "*MixerSoloButton" style:highest "mixer_solo_button"
widget "*MixerSoloButton-alternate" style:highest "mixer_solo_button_alternate"
widget "*MixerSoloButton-alternate2" style:highest "mixer_solo_button_alternate2"
widget "*MixerSoloButton-active" style:highest "mixer_solo_button_active"
widget "*TrackLoopButton*" style:highest "track_loop_button"
widget "*PanAutomationLineSelector*" style:highest "multiline_combo"
widget "*EditorRulerLabel*" style:highest "ruler_label"
widget "*MixerInvertButton*" style:highest "mixer_invert_button"
widget "*MixerInvertButton.*" style:highest "mixer_invert_button"
widget "*MixerAutomationRecordingButton*" style:highest "very_small_button"
widget "*MixerAutomationRecordingButton.*" style:highest "very_small_button"
widget "*MixerAutomationModeButton*" style:highest "very_small_button"
widget "*MixerAutomationModeButton.*" style:highest "very_small_button"
widget "*MixerAutomationPlaybackButton*" style:highest "very_small_button"
widget "*MixerAutomationPlaybackButton.*" style:highest "very_small_button"
widget "*MixerNameButton" style:highest "very_small_button"
widget "*MixerNameButtonLabel" style:highest "very_small_button"
widget "*MixerGroupButton" style:highest "very_small_button"
widget "*MixerGroupButtonLabel" style:highest "very_small_button"
widget "*MixerCommentButton" style:highest "very_small_button"
widget "*MixerCommentButton*" style:highest "very_small_button"
widget "*MixerMonoButton*" style:highest "very_small_button"
widget "*EditGroupButton" style:highest "very_small_button"
widget "*EditGroupButtonLabel" style:highest "very_small_button"
widget "*TransportButton" style:highest "transport_button"
widget "*TransportButton-active" style:highest "transport_button_active"
widget "*TransportButton-alternate2" style:highest "transport_button_alternate_two"
widget "*TransportSyncButton" style:highest "transport_button"
widget "*TransportSyncButton*" style:highest "transport_button"
widget "*TransportSyncButton-active" style:highest "sync_alert"
widget "*TransportSyncButton-active*" style:highest "sync_alert"
widget "*ShuttleButton" style:highest "transport_button"
widget "*ShuttleButton*" style:highest "transport_button"
widget "*ShuttleDisplay" style:highest "transport_button"
widget "*ShuttleDisplay*" style:highest "transport_button"
widget "*ShuttleControl" style:highest "shuttle_control"
widget "*TransportRecButton" style:highest "transport_rec_button"
widget "*TransportRecButton*" style:highest "transport_rec_button"
widget "*TransportRecButton-active" style:highest "transport_rec_button_active"
widget "*TransportRecButton-active*" style:highest "transport_rec_button_active"
widget "*TransportRecButton-alternate" style:highest "transport_rec_button_alternate"
widget "*TransportRecButton-alternate*" style:highest "transport_rec_button_alternate"
widget "*TransportRecButton*" style:highest "transport_rec_button"
widget "*RecordingXrunWarningWindow" style:highest "xrun_warn"
widget "*RecordingXrunWarningWindow*" style:highest "xrun_warn"
#widget "*MainMenuBar" style:highest "menu_bar_base"
widget "*ErrorMessage" style:highest "error_message"
widget "*FatalMessage" style:highest "fatal_message"
widget "*InfoMessage" style:highest "info_message"
widget "*WarningMessage" style:highest "warning_message"

widget "*big clock" style:highest "monospace massive_text"
widget "*selection clock" style:highest "small_mono_text"
widget "*punch clock" style:highest "small_mono_text"
widget "*transport clock" style:highest "bigger_mono_text"
widget "*secondary clock" style:highest "bigger_mono_text"
widget "*transport delta clock" style:highest "bigger_mono_text"
widget "*secondary delta clock" style:highest "bigger_mono_text"
widget "*silence duration clock" style:highest "medium_text"
widget "*edit point clock" style:highest "medium_text"
widget "*nudge clock" style:highest "small_mono_text"
widget "*menubar clock" style:highest "medium_text"
widget "*zoomrange clock" style:highest "medium_text"
widget "*timecodeoffset clock" style:highest "medium_text"

widget "*TransportLabel" style:highest "small_bold_text"
widget "*TakeLabel" style:highest "small_bold_text"
widget "*LocationLabel" style:highest "small_bold_text"
widget "*WipeLabel" style:highest "small_bold_text"
widget "*TakeTagLabel" style:highest "small_bold_text"
widget "*ToolBarLabel" style:highest "small_bold_text"
widget "*EditorDisplayLabel" style:highest "small_bold_text"
widget "*NewSessionLabel" style:highest "larger_text"
widget "*GlobalButtonLabel" style:highest "default_generic"
widget "*ClickButton" style:highest "medium_entry"
widget "*RegionNameDisplay" style:highest "medium_entry"
widget "*PluginDisplay" style:highest "medium_entry"
widget "*SelectionDisplay" style:highest "medium_entry"
widget "*HistorySelector" style:highest "medium_entry"
widget "*LocationSelector" style:highest "medium_entry"
widget "*TakeSelector" style:highest "medium_entry"
widget "*RegionSelector" style:highest "medium_entry"
widget "*TimecodeRuler" style:highest "editor_time_ruler"
widget "*BBTRuler" style:highest "editor_time_ruler"
widget "*SamplesRuler" style:highest "editor_time_ruler"
widget "*TimecodeRuler" style:highest "editor_time_ruler"
widget "*FramesRuler" style:highest "editor_time_ruler"
widget "*MinSecRuler" style:highest "editor_time_ruler"
widget "*BaseFrame" style:highest "base_frame"

widget "*SendStripBase" style:highest "send_strip_base"
widget "*AudioTrackStripBase" style:highest "audio_track_base"
widget "*AudioBusStripBase" style:highest "audio_bus_base"
widget "*MidiTrackStripBase" style:highest "midi_track_base"
widget "*AudioTrackStripBaseInactive" style:highest "track_controls_inactive"
widget "*AudioBusStripBaseInactive" style:highest "track_controls_inactive"
widget "*MidiTrackStripBaseInactive" style:highest "track_controls_inactive"
widget "*FaderMetricsStrip" style:highest "audio_track_metrics"
widget "*AudioTrackMetrics" style:highest "audio_track_metrics"
widget "*MidiTrackMetrics" style:highest "midi_track_metrics"
widget "*AudioBusMetrics" style:highest "audio_bus_metrics"
widget "*AudioMidiTrackMetrics" style:highest "midi_track_metrics"
widget "*AudioTrackMetricsLeft" style:highest "audio_track_metrics"
widget "*MidiTrackMetricsLeft" style:highest "midi_track_metrics"
widget "*AudioBusMetricsLeft" style:highest "audio_bus_metrics"
widget "*AudioMidiTrackMetricsLeft" style:highest "midi_track_metrics"
widget "*AudioTrackMetricsRight" style:highest "audio_track_metrics"
widget "*MidiTrackMetricsRight" style:highest "midi_track_metrics"
widget "*AudioBusMetricsRight" style:highest "audio_bus_metrics"
widget "*AudioMidiTrackMetricsRight" style:highest "midi_track_metrics"
widget "*AudioTrackMetricsInactive" style:highest "audio_track_metrics_inactive"
widget "*MidiTrackMetricsInactive" style:highest "midi_track_metrics_inactive"
widget "*AudioBusMetricsInactive" style:highest "audio_bus_metrics_inactive"
widget "*AudioMidiTrackMetricsInactive" style:highest "midi_track_metrics_inactive"

widget "*TimeAxisViewControlsBaseUnselected" style:highest "audio_track_base"
widget "*AudioTrackControlsBaseUnselected" style:highest "audio_track_base"
widget "*MidiTrackControlsBaseUnselected" style:highest "midi_track_base"

widget "*SendUIFader" style:highest "plugin_slider"
widget "*ReturnUIFader" style:highest "plugin_slider"

widget "*AudioTrackFader" style:highest "audio_track_fader"
widget "*MidiTrackFader" style:highest "midi_track_fader"
widget "*AudioBusFader" style:highest "audio_bus_fader"

widget "*BusControlsBaseUnselected" style:highest "audio_bus_base"
widget "*TrackSeparator" style:highest "track_separator"
widget "*TrackEditIndicator0*" style:highest "edit_group_0"
widget "*TrackEditIndicator1*" style:highest "edit_group_1"
widget "*TrackEditIndicator2*" style:highest "edit_group_2"
widget "*TrackEditIndicator3*" style:highest "edit_group_3"
widget "*TrackEditIndicator4*" style:highest "edit_group_3"
widget "*TrackEditIndicator5*" style:highest "edit_group_3"
widget "*TrackEditIndicator6*" style:highest "edit_group_3"
widget "*TrackEditIndicator7*" style:highest "edit_group_3"
widget "*EditorTrackNameDisplay" style:highest "track_name_display"
widget "*EditorTrackNameDisplay*" style:highest "track_name_display"
widget "*CrossfadeEditAuditionButton" style:highest "bright_when_active"
widget "*CrossfadeEditAuditionButton*" style:highest "bright_when_active"
widget "*CrossfadeEditCurveButton" style:highest "bright_when_active"
widget "*CrossfadeEditCurveButton*" style:highest "bright_when_active"
widget "*CrossfadeEditLabel" style:highest "medium_text"
widget "*CrossfadeEditFrame" style:highest "base_frame"
widget "*MouseModeButton" style:highest "mouse_mode_button"
widget "*MouseModeButton*" style:highest "default_button"
widget "*EditorMainCanvas" style:highest "main_canvas_area"
widget "*AudioTrackControlsBaseInactiveUnselected" style:highest "track_controls_inactive"
widget "*BusControlsBaseInactiveUnselected" style:highest "track_controls_inactive"
widget "*AutomationTrackControlsBaseInactiveUnselected" style:highest "track_controls_inactive"
widget "*AudioTrackControlsBaseInactiveSelected" style:highest "track_controls_inactive"
widget "*BusControlsBaseInactiveSelected" style:highest "track_controls_inactive"
widget "*AutomationTrackControlsBaseInactiveSelected" style:highest "track_controls_inactive"
widget "*AudioTrackControlsBaseSelected" style:highest "edit_controls_base_selected"
widget "*MidiTrackControlsBaseSelected" style:highest "edit_controls_base_selected"
widget "*BusControlsBaseSelected" style:highest "edit_controls_base_selected"
widget "*AutomationTrackControlsBase" style:highest "automation_track_controls_base"
widget "*AutomationTrackControlsBaseSelected" style:highest "edit_controls_base_selected"
#widget "*EditorMenuBar*" style:highest "black_mackie_menu_bar"
#widget "*MainMenuBar*" style:highest "black_mackie_menu_bar"
widget "*ZoomClickBox" style:highest "medium_bold_entry"
widget "*PluginParameterLabel" style:highest "medium_text"
widget "*PluginNameInfo" style:highest "plugin_name_text"
widget "*PluginMakerInfo" style:highest "plugin_maker_text"
widget "*PluginParameterInfo" style:highest "medium_text"
widget "*MotionControllerValue" style:highest "small_entry"
widget "*MotionControllerValue*" style:highest "small_entry"
widget "*ParameterValueDisplay" style:highest "medium_bold_entry"
widget "*PluginUIClickBox" style:highest "medium_bold_entry"
widget "*PluginUIClickBox*" style:highest "medium_bold_entry"
widget "*ProcessorControlSlider" style:highest "processor_control_slider"
widget "*processor control button" style:highest "processor_control_button"
widget "*PluginSlider" style:highest "plugin_slider"
widget "*GainFader" style:highest "plugin_slider"
widget "*MixerTrackCommentArea" style:highest "option_entry"
widget "*MixerPanZone" style:highest "pan_zone"
widget "*LHSList" style:highest "treeview_display"
widget "*RegionEditorLabel" style:highest "medium_text"
widget "*RegionEditorSmallLabel" style:highest "small_text"
widget "*RegionEditorEntry" style:highest "medium_entry"
widget "*RegionEditorToggleButton" style:highest "paler_bright_when_active"
widget "*RegionEditorToggleButton*" style:highest "paler_bright_when_active"
widget "*MixerStripSpeedBase" style:highest "small_entry"
widget "*MixerStripSpeedBase*" style:highest "small_entry"
widget "*MixerStripSpeedBaseNotOne" style:highest "small_red_on_black_entry"
widget "*MixerStripSpeedBaseNotOne*" style:highest "small_red_on_black_entry"
widget "*MixerStripGainDisplay" style:highest "small_entry"
widget "*MixerStripGainDisplay*" style:highest "small_entry"
widget "*MixerStripGainUnitButton" style:highest "very_small_button"
widget "*MixerStripGainUnitButton*" style:highest "very_small_button"
widget "*MixerStripMeterPreButton" style:highest "very_small_button"
widget "*MixerStripMeterPreButton*" style:highest "very_small_button"
widget "*MixerStripPeakDisplay" style:highest "red_active_small_entry"
widget "*MixerStripPeakDisplay*" style:highest "red_active_small_entry"
widget "*MixerStripPeakDisplayPeak*" style:highest "peak_display_peaked_entry"
widget "*MixerStripSelectedFrame" style:highest "selected_strip_frame"
widget "*MixerStripFrame" style:highest "base_frame"
widget "*HWMonitorButton" style:highest "bright_when_active"
widget "*HWMonitorButton*" style:highest "bright_when_active"
widget "*monitor section invert" style:highest "monitor_invert_button"
widget "*monitor section cut" style:highest "monitor_mute_button"
widget "*monitor section dim" style:highest "monitor_dim_button"
widget "*monitor section mono" style:highest "monitor_mono_button"
widget "*monitor solo exclusive" style:highest "monitor_mono_button"
widget "*monitor solo override" style:highest "monitor_mono_button"
widget "*MonitorIsoAlert" style:highest "monitor_iso_button"
widget "*MonitorIsoAlert*" style:highest "monitor_iso_button"
widget "*BypassButton" style:highest "bright_when_active"
widget "*BypassButton*" style:highest "bright_when_active"
widget "*FadeCurve" style:highest "medium_bold_entry"
widget "*FadeCurve*" style:highest "medium_bold_entry"
widget "*RouteParamsListDisplay" style:highest "inspector_track_list_display"
widget "*TearOffArrow" style:highest "tearoff_arrow"
widget "*RouteParamsTitleButton" style:highest "medium_text"
widget "*RouteParamsTitleLabel" style:highest "medium_text"
widget "*PluginAutomateRecordButton" style:highest "small_red_active_and_selected_button"
widget "*PluginAutomateRecordButton*" style:highest "small_red_active_and_selected_button"
widget "*PluginAutomatePlayButton" style:highest "small_red_active_and_selected_button"
widget "*PluginAutomatePlayButton*" style:highest "small_red_active_and_selected_button"
widget "*PluginAutomateButton" style:highest "small_button"
widget "*PluginAutomateButton*" style:highest "small_button"
widget "*PluginSaveButton" style:highest "small_button"
widget "*PluginSaveButton*" style:highest "small_button"
widget "*PluginEditButton*" style:highest "small_button"
widget "*PluginLoadButton" style:highest "small_button"
widget "*PluginLoadButton*" style:highest "small_button"
widget "*PluginEditorButton" style:highest "default_toggle_button"

widget "*meterstripPPM" style:highest "meter_strip_ppm"
widget "*meterstripDPM" style:highest "meter_strip_dpm"
widget "*meterstripVU" style:highest "meter_strip_vu"
widget "*BlackSeparator" style:highest "meter_strip_sep"

widget "*MetricDialogFrame" style:highest "base_frame"
widget "*MetricEntry" style:highest "medium_bold_entry"
widget "*MetricButton" style:highest "default_button"
widget "*MetricButton.*" style:highest "default_button"
widget "*MetricLabel" style:highest "medium_text"
widget "*TimeStretchButton" style:highest "default_button"
widget "*TimeStretchButton.*" style:highest "default_button"
widget "*ChoiceWindow" style:highest "default_generic"
widget "*ChoicePrompt" style:highest "default_generic"
widget "*ChoiceButton" style:highest "default_button"
widget "*ChoiceButton*" style:highest "default_button"
widget "*SelectionModeButton" style:highest "default_button"
widget "*SelectionModeButton*" style:highest "default_button"
widget "*TrackLabel" style:highest "medium_text"
widget "*TrackPlugName" style:highest "medium_text"
widget "*TrackParameterName" style:highest "small_italic_text"
widget "*AddRouteDialog*" style:highest "medium_text"
widget "*AddRouteDialog.GtkLabel" style:highest "medium_text"
widget "*AddRouteDialogChannelChoice" style:highest "medium_bold_entry"
widget "*AddRouteDialogSpinner" style:highest "medium_bold_entry"
widget "*AddRouteDialogSpinner*" style:highest "medium_bold_entry"
widget "*AddRouteDialogRadioButton" style:highest "bright_when_active"
widget "*AddRouteDialogButton" style:highest "default_button"
widget "*AddRouteDialogNameTemplateEntry" style:highest "medium_bold_entry"
widget "*NewSessionIOLabel" style:highest "large_bold_text"
widget "*NewSessionSR1Label" style:highest "red_medium_text"
widget "*NewSessionSR2Label" style:highest "medium_text"
widget "*NewSessionChannelChoice" style:highest "medium_bold_entry"
widget "*NewSessionMainButton" style:highest "large_bold_text"
widget "*NewSessionMainButton*" style:highest "large_bold_text"
widget "*NewSessionMainLabel" style:highest "large_bold_text"
widget "*location edit row clock" style:highest "medium_text"
widget "*LocationEditNameLabel" style:highest "medium_text"
widget "*LocationEditSetButton" style:highest "location_row_button"
widget "*LocationEditSetButton*" style:highest "location_row_button"
widget "*LocationEditGoButton" style:highest "location_row_button"
widget "*LocationEditGoButton*" style:highest "location_row_button"
widget "*LocationEditCdButton" style:highest "small_red_active_and_selected_button"
widget "*LocationEditCdButton*" style:highest "small_red_active_and_selected_button"
widget "*LocationEditHideButton" style:highest "small_red_active_and_selected_button"
widget "*LocationEditHideButton*" style:highest "small_red_active_and_selected_button"
widget "*LocationEditLockButton*" style:highest "small_red_active_and_selected_button"
widget "*LocationEditGlueButton*" style:highest "small_red_active_and_selected_button"
widget "*LocationEditNumberLabel" style:highest "small_text"
widget "*LocationLocEditorFrame" style:highest "base_frame"
widget "*LocationRangeEditorFrame" style:highest "base_frame"
widget "*LocationEditNameEntry" style:highest "option_entry"
widget "*LocationAddLocationButton" style:highest "default_button"
widget "*LocationAddLocationButton*" style:highest "default_button"
widget "*LocationAddRangeButton" style:highest "default_button"
widget "*LocationAddRangeButton*" style:highest "default_button"
widget "*LocationEditRemoveButton" style:highest "location_row_button"
widget "*LocationEditRemoveButton*" style:highest "location_row_button"
widget "*ChannelCountSelector" style:highest "medium_bold_entry"
widget "*ChannelCountSelector.GtkArrow" style:highest "default_generic"
widget "*RegionListWholeFile" style:highest "treeview_parent_node"
widget "*EditorHScrollbar" style:highest "editor_hscrollbar"
widget "*MidiListView*" style:highest "treeview_display"
widget "*ProcessorList*" style:highest "processor_list"
widget "*PortMatrixLabel*" style:highest "small_text"
widget "*MidiTracerTextView" style:highest "midi_tracer_textview"
widget "*solo isolate" style:highest "solo_isolate"
widget "*meterbridge label" style:highest "meterbridge_label"
widget "*solo safe" style:highest "solo_safe"
widget "*ContrastingPopup" style:highest "contrasting_popup"
widget "*ContrastingPopup*" style:highest "contrasting_popup"
widget "*MidiChannelSelectorButton" style:highest "midi_channel_selector_button"
widget "*TimeInfoSelectionTitle" style:highest "very_small_text"
widget "*TimeInfoSelectionLabel" style:highest "very_small_text"
widget "*TimeInfoPunchTitle" style:highest "very_small_text"
widget "*TimeInfoPunchButton" style:highest "very_small_text"
widget "*TimeInfoBox" style:highest "time_info_box"
widget "*StatusBarBox" style:highest "status_bar_box"
widget "*RouteNameEditorEntry" style:highest "text_cell_entry"
widget "*RegionNameEditorEntry" style:highest "text_cell_entry"
widget "*EditorRouteGroupsAllGroupButton" style:highest "default_toggle_button"
widget "*MidiSoundNotesButton" style:highest "default_toggle_button"
widget "*MeasureLatencyButton" style:highest "default_toggle_button"
widget "*monitor" style:highest "monitor"
widget "*processor prefader" style:highest "processor"
widget "*processor fader" style:highest "processor"
widget "*processor postfader" style:highest "processor"
widget "*MonitorSectionLabel" style:highest "small_text"
widget "*mute button" style:highest "monitor"
widget "*invert button" style:highest "monitor"
widget "*send alert button" style:highest "small_text"
widget "*solo button" style:highest "monitor"
widget "*rude solo" style:highest "small_text"
widget "*rude audition" style:highest "small_text"
widget "*feedback alert" style:highest "small_text"
widget "*transport option button" style:highest "small_text"
widget "*transport active option button" style:highest "small_text"
widget "*plugin bypass button" style:highest "small_text"
widget "*punch button" style:highest "small_text"
