/*
    Copyright (C) 2010 Paul Davis

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef __pbd_semutils_h__
#define __pbd_semutils_h__

#include <semaphore.h>

namespace PBD {

class ProcessSemaphore {
  private:
#ifdef __APPLE__
	sem_t* _sem;
	sem_t* ptr_to_sem() const { return _sem; }
#else
	mutable sem_t _sem;
	sem_t* ptr_to_sem() const { return &_sem; }
#endif

  public:
	ProcessSemaphore (const char* name, int val);
	~ProcessSemaphore ();

	int signal () { return sem_post (ptr_to_sem()); }
	int wait () { return sem_wait (ptr_to_sem()); }
};

}

#endif /* __pbd_semutils_h__ */
