#include "evoral/types.hpp"

uint64_t PBD::DEBUG::Sequence = PBD::new_debug_bit ("sequence");
uint64_t PBD::DEBUG::Note = PBD::new_debug_bit ("note");
uint64_t PBD::DEBUG::ControlList = PBD::new_debug_bit ("controllist");

