#ifndef __libpbd_version_h__
#define __libpbd_version_h__
extern const char* libpbd_revision;
extern int libpbd_major_version;
extern int libpbd_minor_version;
extern int libpbd_micro_version;
#endif /* __libpbd_version_h__ */
